# Copyright 2014-2016 Neil Freeman contact@fakeisthenewreal.org
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from twitter_bot_utils import helpers
import wordfilter


def generator(tweets, return_status=None, **kwargs):
    '''
    Returns a generator that returned a filtered input iterable of tweets or
    tweet-like objects (tweepy.Status objects or dictionaries).

    Args:
        tweets (iterable):
        return_status (boolean): If true, returns entire status with modified test
        no_retweets (boolean): Exclude retweets (e.g. strings beginning RT) (default False)
        no_replies (boolean): Exclude replies (e.g. strings beginning @screen_name) (default False)
        no_mentions (boolean): Filter out mentions (e.g. strings containing @screen_name) (default False)
        no_badwords (boolean): Exclude derogatory terms for people (default True)
        no_urls (boolean): filter out exclude urls (default False)
        no_hashtags (boolean): filter out hashtags (default False)
        no_media (boolean): filter out media (twitter objects only) (default False)
        no_symbols (boolean): filter out symbols (twitter objects only) (default False)
    '''

    tweet_checker = construct_tweet_checker(
        no_retweets=kwargs.get('no_retweets'),
        no_replies=kwargs.get('no_replies'),
        no_badwords=kwargs.get('no_replies', True)
    )

    tweet_filter = construct_tweet_filter(
        no_mentions=kwargs.get('no_mentions'),
        no_urls=kwargs.get('no_urls'),
        no_media=kwargs.get('no_media'),
        no_hashtags=kwargs.get('no_hashtags'),
        no_symbols=kwargs.get('no_symbols')
    )

    for status in tweets:
        if not tweet_checker(status):
            continue

        text = helpers.format_text(tweet_filter(status))

        if return_status:
            try:
                status.text = text
            except AttributeError:
                status['text'] = text

            yield status

        yield text


def isreply(tweet):
    '''
    Checks if a given tweet is a reply.
    If tweet is a string, returns True when text starts with '@'.
    If tweet is a tweepy.Status object, returns True when ``in_reply_to_status_id`` or
    ``in_reply_to_user_id`` is set.

    Args:
        tweet (str/tweepy.Status/dict): A string, tweepy.Status object, or Status-like dict.

    Returns:
        bool
    '''
    try:
        return bool(tweet.in_reply_to_user_id)

    except AttributeError:
        try:
            return bool(tweet.get('in_reply_to_user_id') or tweet.get('in_reply_to_status_id'))

        except AttributeError:
            try:
                return tweet[0] == "@"

            except AttributeError:
                pass

    return False


def isretweet(tweet):
    '''
    Checks if a given tweet is a retweet.
    If tweet is a string, returns True when text starts with 'RT '.
    If tweet is a tweepy.Status object, returns True when ``retweeted_status`` or
    ``retweeted_status_id`` is set.

    Args:
        tweet (str/tweepy.Status/dict): A string, tweepy.Status object, or Status-like dict.

    Returns:
        bool
    '''
    try:
        return bool(tweet.retweeted)

    except AttributeError:
        try:
            return bool(tweet.get('retweeted_status') or tweet.get('retweeted_status_id'))

        except AttributeError:
            try:
                return tweet[:3].upper() == 'RT '

            except AttributeError:
                pass

    return False


def isblacklisted(tweet):
    '''
    Checks if a given tweet contains a word blacklisted by WordFilter.

    Args:
        tweet (str/tweepy.Status/dict): A string, tweepy.Status object, or Status-like dict.

    Returns:
        bool
    '''
    try:
        return wordfilter.blacklisted(tweet.text)

    except AttributeError:
        try:
            return wordfilter.blacklisted(tweet['text'])

        except (KeyError, TypeError):
            return wordfilter.blacklisted(tweet)

    return False


def construct_tweet_checker(no_retweets=False, no_replies=False, no_badwords=True):
    '''
    Returns a tweet checker, a function that checks if tweets pass the tests.

    Args:
        no_retweets (boolean): Checker filters out retweets (default: False).
        no_replies  (boolean): Checker filters out replies (default: False).
        no_badwords  (boolean): Checker filters out blacklisted words (default: True).

    Returns:
        function
    '''
    checks = []

    if no_retweets:
        checks.append(isretweet)

    if no_replies:
        checks.append(isreply)

    if no_badwords:
        checks.append(isblacklisted)

    def checker(tweet):
        return not any(isbad(tweet) for isbad in checks)

    return checker


def construct_tweet_filter(no_mentions=False, no_urls=False, no_media=False, no_hashtags=False, no_symbols=False):
    '''
    Returns a filter function for tweet text.

    Args:
        no_mentions (boolean): filter out mentions.
        no_urls (boolean): filter out urls.
        no_media (boolean): filter out media.
        no_hashtags (boolean): filter out hashtags.
        no_symbols (boolean): filter out symbols.

    Returns:
        function
    '''
    entitytypes = []

    if no_mentions:
        entitytypes.append('user_mentions')

    if no_hashtags:
        entitytypes.append('hashtags')

    if no_urls:
        entitytypes.append('urls')

    if no_media:
        entitytypes.append('media')

    if no_symbols:
        entitytypes.append('symbols')

    def filterer(tweet):
        # ignore strings
        try:
            text = helpers.remove_entities(tweet, entitytypes)
        except AttributeError:
            text = tweet

        # Older tweets don't have entities
        if no_urls:
            # regex stolen from http://stackoverflow.com/questions/6883049/regex-to-find-urls-in-string-in-python
            text = re.sub(r"http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+", '', text)

        if no_mentions:
            text = re.sub(r'@\w+', '', text)

        if no_hashtags:
            text = re.sub(r'#\w+', '', text)

        if no_symbols:
            text = re.sub(r'\$[a-zA-Z]+', '', text)

        return text

    return filterer



# twitter_markov - Create markov chain ("_ebooks") accounts on Twitter
# Copyright 2014-2016 Neil Freeman contact@fakeisthenewreal.org

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .twitter_markov import TwitterMarkov
from . import checking

__version__ = "0.5.0"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2014-2016 Neil Freeman contact@fakeisthenewreal.org
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, print_function
import os
import sys
from signal import signal, SIGPIPE, SIG_DFL
import argparse
import twitter_bot_utils as tbu
from . import TwitterMarkov
from . import checking
from . import __version__ as version


TWEETER_DESC = 'Post markov chain ("ebooks") tweets to Twitter'
LEARNER_DESC = 'Turn a twitter archive into a twitter_markov-ready text file'


def main():
    parser = argparse.ArgumentParser(
        'twittermarkov', description='Tweet with a markov bot, or teach it from a twitter archive.')

    tbu.args.add_default_args(parser, version, ())

    subparsers = parser.add_subparsers()

    tweeter = subparsers.add_parser('tweet', description=TWEETER_DESC, usage='%(prog)s [options]')
    tbu.args.add_default_args(tweeter, include=('user', 'config', 'dry-run', 'verbose', 'quiet'))
    tweeter.add_argument('-r', '--reply', action='store_const', const='reply',
                         dest='action', help='tweet responses to recent mentions')
    tweeter.add_argument('--corpus', dest='corpus', metavar='corpus', type=str,
                         help='text file, one sentence per line')
    tweeter.add_argument('--max-len', type=int, default=140, help='maximum output length. default: 140')
    tweeter.add_argument('--state-size', type=int, help='model state size. default: 2')
    tweeter.add_argument('--no-learn', dest='learn', action='store_false',
                         help='skip learning (by default, recent tweets from the "parent" account are added to corpus)')
    tweeter.set_defaults(subparser='tweet', func=tweet_func, action='tweet')

    learner = subparsers.add_parser('corpus', description=LEARNER_DESC, usage="%(prog)s [options] archive corpus")
    learner.add_argument('-o', type=str, dest='output', metavar='corpus',
                         help='output text file (defaults to stdout)', default='/dev/stdout')
    learner.add_argument('--no-retweets', action='store_true', help='skip retweets')
    learner.add_argument('--no-replies', action='store_true', help='filter out replies')
    learner.add_argument('--no-mentions', action='store_true', help='filter out mentions')
    learner.add_argument('--no-urls', action='store_true', help='filter out urls')
    learner.add_argument('--no-media', action='store_true', help='filter out media')
    learner.add_argument('--no-hashtags', action='store_true', help='filter out hashtags')
    learner.add_argument('-q', '--quiet', action='store_true', help='run quietly')
    learner.add_argument('archive', type=str, metavar='archive',
                         default=os.getcwd(), help='archive csv file (e.g. tweets.csv found in Twitter archive)')

    learner.set_defaults(subparser='learn', func=learn_func, action='learn')

    args = parser.parse_args()
    try:
        func = args.func
    except AttributeError:
        parser.parse_args(['--help'])

    argdict = vars(args)
    del argdict['func']

    if args.subparser == 'tweet':
        func(**argdict)

    elif args.subparser == 'learn':
        func(**argdict)


def tweet_func(action, max_len=None, **kwargs):
    tm = TwitterMarkov(**kwargs)
    try:
        if action == 'tweet':
            tm.log.debug('tweeting')
            tm.tweet(max_len=max_len)

        elif action == 'reply':
            tm.log.debug('replying')
            tm.reply_all(max_len=max_len)

    except RuntimeError:
        tm.log.error('model was unable to compose a tweet')
        return


def learn_func(**kwargs):
    if not kwargs['quiet']:
        print("Reading " + kwargs['archive'], file=sys.stderr)

    archive = tbu.archive.read_csv(kwargs.get('archive'))
    gen = checking.generator(archive, **kwargs)
    tweets = (tweet.replace(u'\n', u' ') + '\n' for tweet in gen)

    if kwargs['output'] in ('-', '/dev/stdout'):
        signal(SIGPIPE, SIG_DFL)
        sys.stdout.writelines(tweets)

    else:
        if not kwargs['quiet']:
            print("Writing " + kwargs['output'], file=sys.stderr)

        with open(kwargs.get('output'), 'w') as f:
            f.writelines(tweets)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright 2014-2016 Neil Freeman contact@fakeisthenewreal.org
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals, print_function
import os
import re
import logging
from collections import Iterable
import Levenshtein
import six
import markovify.text
import twitter_bot_utils as tbu
from wordfilter import Wordfilter
from . import checking

LEVENSHTEIN_LIMIT = 0.70


class TwitterMarkov(object):
    """
    Posts markov-generated text to twitter

    Args:
        screen_name (str): Twitter user account
        corpus (str): Text file to read to generate text.
        api (:ref:`tweepy.API <tweepy:tweepy.api>`): API to use to post tweets.
        dry_run (boolean): If set, TwitterMarkov won't actually post tweets.
        blacklist (Sequence): A list of words to avoid generating.
    """

    default_model = None
    _recently_tweeted = []

    def __init__(self, screen_name, corpus=None, **kwargs):
        if 'api' in kwargs:
            self.api = kwargs.pop('api')
        else:
            self.api = tbu.API(screen_name=screen_name, **kwargs)

        try:
            self.log = self.api.logger
        except AttributeError:
            self.log = logging.getLogger(screen_name)

        self.screen_name = screen_name
        self.config = self.api.config
        self.dry_run = kwargs.pop('dry_run', False)

        self.log.debug('screen name: %s', screen_name)
        self.log.debug("dry run: %s", self.dry_run)

        try:
            corpus = corpus or self.config.get('corpus')

            if isinstance(corpus, six.string_types):
                corpora = [corpus]

            elif isinstance(corpus, Iterable):
                corpora = corpus

            else:
                raise RuntimeError('Unable to find any corpora!')

            self.corpora = [b for b in corpora if b is not None]

            state_size = kwargs.get('state_size', self.config.get('state_size'))

            self.models = self._setup_models(self.corpora, state_size)

        except RuntimeError as e:
            self.log.error(e)
            raise e

        self.log.debug('models: %s', list(self.models.keys()))

        blacklist = kwargs.get('blacklist') or self.config.get('blacklist', [])
        self.wordfilter = Wordfilter()
        self.wordfilter.add_words(blacklist)

        self.log.debug('blacklist: %s terms', len(self.wordfilter.blacklist))

        if kwargs.get('learn', True):
            self.log.debug('learning...')
            self.learn_parent()

    def _setup_models(self, corpora, state_size):
        """
        Given a list of paths to corpus text files, set up markovify models for each.
        These models are returned in a dict, (with the basename as key).
        """
        out = dict()
        state_size = state_size or 2
        self.log.debug('setting up models (state_size=%s)', state_size)

        try:
            for pth in corpora:
                corpus_path = os.path.expanduser(pth)
                name = os.path.basename(corpus_path)

                with open(corpus_path) as m:
                    out[name] = markovify.text.NewlineText(m.read(), state_size=state_size)

        except AttributeError as e:
            self.log.error(e)
            self.log.error("Probably couldn't find the model file.")
            raise e

        except IOError as e:
            self.log.error(e)
            self.log.error('Error reading %s', corpus_path)
            raise e

        self.default_model = os.path.basename(corpora[0])

        return out

    @property
    def recently_tweeted(self):
        '''Returns recent tweets from ``self.screen_name``.'''
        if len(self._recently_tweeted) == 0:
            recent_tweets = self.api.user_timeline(self.screen_name, count=self.config.get('checkback', 20))
            self._recently_tweeted = [x.text for x in recent_tweets]

        return self._recently_tweeted

    def check_tweet(self, text):
        '''Check if a string contains blacklisted words or is similar to a recent tweet.'''
        text = text.strip().lower()

        if len(text) == 0:
            self.log.info("Rejected (empty)")
            return False

        if self.wordfilter.blacklisted(text):
            self.log.info("Rejected (blacklisted)")
            return False

        if tbu.helpers.length(text) > 280:
            self.log.info("Rejected (too long)")
            return False

        for line in self.recently_tweeted:
            if text in line.strip().lower():
                self.log.info("Rejected (Identical)")
                return False

            if Levenshtein.ratio(re.sub(r'\W+', '', text), re.sub(r'\W+', '', line.lower())) >= LEVENSHTEIN_LIMIT:
                self.log.info("Rejected (Levenshtein.ratio)")
                return False

        return True

    def reply_all(self, model=None, **kwargs):
        '''Reply to all mentions since the last time ``self.screen_name`` sent a reply tweet.'''
        mentions = self.api.mentions_timeline(since_id=self.api.last_reply)
        self.log.info('replying to all...')
        self.log.debug('mentions found: %d', len(mentions))

        if not self.dry_run:
            for status in mentions:
                self.reply(status, model, **kwargs)

    def reply(self, status, model=None, max_len=140, **kwargs):
        '''
        Compose a reply to the given ``tweepy.Status``.

        Args:
            status (tweepy.Status): status to reply to.
            model (str): name of model.
            max_len (int): maximum length of tweet (default: 140)
        '''
        self.log.debug('Replying to a mention')

        if status.user.screen_name == self.screen_name:
            self.log.debug('Not replying to self')
            return

        if self.wordfilter.blacklisted(status.text):
            self.log.debug('Not replying to tweet with a blacklisted word (%d)', status.id)
            return

        text = self.compose(model, max_len=max_len - 2 - len(status.user.screen_name), **kwargs)
        reply = '@{} {}'.format(status.user.screen_name, text)

        self.log.info(reply)
        self._update(reply, in_reply=status.id_str)

    def tweet(self, model=None, **kwargs):
        '''
        Post a tweet composed by "model" (or the default model).
        Most of these arguments are passed on to Markovify.

        Args:
            model (str): one of self.models
            max_len (int): maximum length of the output (default: 140).
            init_state (tuple): tuple of words to seed the model
            tries (int): (default: 10)
            max_overlap_ratio (float): Used for testing output (default: 0.7).
            max_overlap_total (int): Used for testing output (default: 15)
        '''
        model = self.models[model or self.default_model]
        text = self.compose(model, **kwargs)
        if text:
            self._update(text)

    def _update(self, tweet, in_reply=None):
        if not self.dry_run:
            self.api.update_status(status=tweet, in_reply_to_status_id=in_reply)

    def compose(self, model=None, max_len=140, **kwargs):
        '''
        Returns a string generated from "model" (or the default model).
        Most of these arguments are passed on to Markovify.

        Args:
            model (str): one of self.models
            max_len (int): maximum length of the output (max: 280, default: 140).
            init_state (tuple): tuple of words to seed the model
            tries (int): (default: 10)
            max_overlap_ratio (float): Used for testing output (default: 0.7).
            max_overlap_total (int): Used for testing output (default: 15)

        Returns:
            str
        '''
        model = model or self.models[self.default_model]
        max_len = min(280, max_len)
        self.log.debug('making sentence, max_len=%s, %s', max_len, kwargs)
        text = model.make_short_sentence(max_len, **kwargs)

        if text is None:
            self.log.error('model failed to generate a sentence')
            raise RuntimeError('model failed to generate a sentence')

        # convert to unicode in Python 2
        if hasattr(text, 'decode'):
            text = text.decode('utf8')

        else:
            # Check tweet against blacklist and recent tweets
            if not self.check_tweet(text):
                # checked out: break and return
                text = self.compose(model=model, max_len=max_len, **kwargs)

        self.log.debug('TwitterMarkov: %s', text)

        return text

    def learn_parent(self, corpus=None, parent=None):
        '''
        Add recent tweets from the parent account (since the last time ``self.screen_name`` tweeted)
        to the corpus. This is subject to the filters described in ``bots.yaml``.
        '''
        parent = parent or self.config.get('parent')
        corpus = corpus or self.corpora[0]

        if not parent or not self.api.last_tweet:
            self.log.debug('Cannot teach: missing parent or tweets')
            return

        tweets = self.api.user_timeline(parent, since_id=self.api.last_tweet)

        try:
            gen = checking.generator(tweets,
                                     no_mentions=self.config.get('filter_mentions'),
                                     no_hashtags=self.config.get('filter_hashtags'),
                                     no_urls=self.config.get('filter_urls'),
                                     no_media=self.config.get('filter_media'),
                                     no_symbols=self.config.get('filter_symbols'),
                                     no_badwords=self.config.get('filter_parent_badwords', True),
                                     no_retweets=self.config.get('no_retweets'),
                                     no_replies=self.config.get('no_replies')
                                    )

            self.log.debug('%s is learning', corpus)

            with open(corpus, 'a') as f:
                f.writelines(tweet + '\n' for tweet in gen)

        except IOError as e:
            self.log.error('Learning failed for %s', corpus)
            self.log.error(e)



#!/usr/bin/env python
# twitter_markov - Create markov chain ("_ebooks") accounts on Twitter
# Copyright 2014-2016 Neil Freeman contact@fakeisthenewreal.org

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup

try:
    readme = open('README.rst', 'r').read()
except IOError:
    readme = ''

with open('twitter_markov/__init__.py') as i:
    version = next(r for r in i.readlines() if '__version__' in r).split('=')[1].strip('"\' \n')

setup(
    name='twitter_markov',

    version=version,

    description='Create markov chain ("_ebooks") accounts on Twitter',

    long_description=readme,

    author='Neil Freeman',

    author_email='contact@fakeisthenewreal.org',

    url='https://github.com/fitnr/twitter_markov',

    packages=['twitter_markov'],

    license='GPLv3',

    entry_points={
        'console_scripts': [
            'twittermarkov=twitter_markov.__main__:main',
        ],
    },

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Operating System :: OS Independent',
    ],

    zip_safe=True,

    install_requires=[
        'twitter_bot_utils>=0.11.6.post1,<0.12',
        'markovify>=0.2.4,<0.4',
        'python-Levenshtein>=0.12.0, <0.13',
        'wordfilter>=0.1.8, <0.2.0'
        'pyyaml',
        'tweepy',
        'six'
    ],

    test_suite='tests',
    tests_require=[
        'setuptools>=17.1',
        'pbr>=0.11,<1.0',
        'mock',
    ],
)
